//
// Copyright (c) 2001-2016 by PDFTron Systems Inc. All Rights Reserved.
//

using System;
using System.Collections.Generic;

using pdftron;
using pdftron.Common;
using pdftron.Filters;
using pdftron.SDF;
using pdftron.PDF;

using XSet = System.Collections.Generic.List<int>;

namespace ElementEditTestCS
{
	/// <summary>
	/// The sample code shows how to edit the page display list and how to modify graphics state 
	/// attributes on existing Elements. In particular the sample program strips all images from 
	/// the page, changes path fill color to red, and changes text fill color to blue. 
	/// </summary>
	class ElementEditTest
	{
		private static pdftron.PDFNetLoader pdfNetLoader = pdftron.PDFNetLoader.Instance();

        static Rect GetGlyphBBox(CharIterator itr, pdftron.PDF.Font font,
            double horiz_scale, double font_sz, double ascent, double descent)
        {
            Rect out_bbox = new Rect();
            out_bbox.x1 = itr.Current().x;
            out_bbox.y1 = itr.Current().y;

            double dx = 0;
            if (font.IsSimple())
            {
                dx = font.GetWidth(itr.Current().char_code) / 1000.0;
                dx *= horiz_scale * font_sz;

                out_bbox.x2 = out_bbox.x1 + dx;
                out_bbox.y1 = itr.Current().y + descent;
                out_bbox.y2 = itr.Current().y + ascent;
            }
            else
            {
                int cid = font.MapToCID(itr.Current().char_code);
                dx = font.GetWidth(cid) / 1000.0;
                dx *= horiz_scale * font_sz;

                out_bbox.x2 = out_bbox.x1 + dx;
                out_bbox.y1 = itr.Current().y + descent;
                out_bbox.y2 = itr.Current().y + ascent;
            }

            return out_bbox;
        }

        static Rect GetBBoxTransfRect(Rect inn, Matrix2D mtx)
        {
            double p1x = inn.x1, p1y = inn.y1, p2x = inn.x2, p2y = inn.y1, p3x
         = inn.x2, p3y = inn.y2, p4x = inn.x1, p4y = inn.y2;
            mtx.Mult(ref p1x, ref p1y);
            mtx.Mult(ref p2x, ref p2y);
            mtx.Mult(ref p3x, ref p3y);
            mtx.Mult(ref p4x, ref p4y);

            return new Rect(Math.Min(Math.Min(Math.Min(p1x, p2x), p3x), p4x),
               Math.Min(Math.Min(Math.Min(p1y, p2y), p3y), p4y),
               Math.Max(Math.Max(Math.Max(p1x, p2x), p3x), p4x),
               Math.Max(Math.Max(Math.Max(p1y, p2y), p3y), p4y));
        }

        static void ProcessElements(ElementReader reader, ElementWriter writer, XSet visited)
		{
            Rect sel_rect = new Rect(124.338000, 481.582000, 515.442000, 527.214000);
            ColorPt textColor = new ColorPt(1.0, 0.0, 0.0);
            Rect bbox = new Rect();

            ElementBuilder builder = new ElementBuilder();

            Element element;
			while ((element = reader.Next()) != null) // Read page contents
			{
                switch (element.GetType())
                {
                    case Element.Type.e_text:
                        {
                            ColorPt cpt = element.GetGState().GetFillColor();
                            element.GetBBox(bbox);

                            if (bbox.x2 <= sel_rect.x1 || bbox.x1 >= sel_rect.x2 || bbox.y1
                      >= sel_rect.y2 || bbox.y2 <= sel_rect.y1)
                            {
                                writer.WriteElement(element);
                            }
                            else if (bbox.x2 <= sel_rect.x2 && bbox.x1 >= sel_rect.x1 &&
                                bbox.y1 >= sel_rect.y1 && bbox.y2 <= sel_rect.y2)
                            {
                                GState gs = element.GetGState();
                                gs.SetFillColorSpace(ColorSpace.CreateDeviceRGB());
                                gs.SetFillColor(textColor);
                                writer.WriteElement(element);
                                gs.SetFillColor(cpt);
                            }
                            else
                            {
                                GState gs = element.GetGState();
                                pdftron.PDF.Font font = gs.GetFont();
                                double font_sz = gs.GetFontSize();
                                Rect bbox1 = font.GetBBox();
                                double bbox_height = bbox1.y2 - bbox1.y1;
                                double decent = font_sz * bbox1.y1 / bbox_height;
                                double ascent = font_sz * bbox1.y2 / bbox_height;
                                double horiz_scale = gs.GetHorizontalScale() / 100.0;
                                Matrix2D tmtx = element.GetTextMatrix();
                                Matrix2D mtx = element.GetCTM() * tmtx;

                                for (CharIterator itr = element.GetCharIterator(); itr.HasNext(); itr.Next())
                                {
                                    {
                                        Rect gbox = GetGlyphBBox(itr, font, horiz_scale, font_sz, ascent, decent);
                                        gbox = GetBBoxTransfRect(gbox, mtx);

                                        Element new_element = builder.CreateTextRun(new String((char)(itr.Current().char_code), 1), font, font_sz);

                                        if (element.HasTextMatrix() && itr == element.GetCharIterator())
                                        {
                                            new_element.SetTextMatrix(element.GetTextMatrix());
                                        }

                                        new_element.SetPosAdjustment(element.GetPosAdjustment());
                                        GState ngs = new_element.GetGState();

                                        int i = 0;

                                        for (GSChangesIterator gsitr = reader.GetChangesIterator(); gsitr.HasNext(); gsitr.Next())
                                        {
                                            GState.GStateAttribute st = gsitr.Current();
                                            switch (gsitr.Current())
                                            {
                                                case GState.GStateAttribute.e_halftone:
                                                    ngs.SetHalftone(gs.GetHalftone());
                                                    break;
                                                case GState.GStateAttribute.e_UCR_funct:
                                                    ngs.SetUCRFunct(gs.GetUCRFunct());
                                                    break;
                                                case GState.GStateAttribute.e_BG_funct:
                                                    ngs.SetBlackGenFunct(gs.GetBlackGenFunct());
                                                    break;
                                                case GState.GStateAttribute.e_transfer_funct:
                                                    ngs.SetTransferFunct(gs.GetTransferFunct());
                                                    break;
                                                case GState.GStateAttribute.e_overprint_mode:
                                                    ngs.SetOverprintMode(gs.GetOverprintMode());
                                                    break;
                                                case GState.GStateAttribute.e_fill_overprint:
                                                    ngs.SetFillOverprint(gs.GetFillOverprint());
                                                    break;
                                                case GState.GStateAttribute.e_stroke_overprint:
                                                    ngs.SetStrokeOverprint(gs.GetStrokeOverprint());
                                                    break;
                                                case GState.GStateAttribute.e_auto_stoke_adjust:
                                                    ngs.SetAutoStrokeAdjust(gs.GetAutoStrokeAdjust());
                                                    break;
                                                case GState.GStateAttribute.e_smoothnes:
                                                    ngs.SetSmoothnessTolerance(gs.GetSmoothnessTolerance());
                                                    break;
                                                case GState.GStateAttribute.e_soft_mask:
                                                    ngs.SetSoftMask(gs.GetSoftMask());
                                                    break;
                                                case GState.GStateAttribute.e_alpha_is_shape:
                                                    ngs.SetAISFlag(gs.GetAISFlag());
                                                    break;
                                                case GState.GStateAttribute.e_opacity_stroke:
                                                    ngs.SetStrokeOpacity(gs.GetStrokeOpacity());
                                                    break;
                                                case GState.GStateAttribute.e_opacity_fill:
                                                    ngs.SetFillOpacity(gs.GetFillOpacity());
                                                    break;
                                                case GState.GStateAttribute.e_blend_mode:
                                                    ngs.SetBlendMode(gs.GetBlendMode());
                                                    break;
                                                case GState.GStateAttribute.e_text_pos_offset:
                                                    //ngs.Settext_pos_offset(gs.Gettext_pos_offset());
                                                    break;
                                                case GState.GStateAttribute.e_text_knockout:
                                                    ngs.SetTextKnockout(gs.IsTextKnockout());
                                                    break;
                                                case GState.GStateAttribute.e_text_rise:
                                                    ngs.SetTextRise(gs.GetTextRise());
                                                    break;
                                                case GState.GStateAttribute.e_text_render_mode:
                                                    ngs.SetTextRenderMode(gs.GetTextRenderMode());
                                                    break;
                                                case GState.GStateAttribute.e_font_size:
                                                    ngs.SetFont(gs.GetFont(), gs.GetFontSize());
                                                    break;
                                                case GState.GStateAttribute.e_font:
                                                    ngs.SetFont(gs.GetFont(), gs.GetFontSize());
                                                    break;
                                                case GState.GStateAttribute.e_leading:
                                                    ngs.SetLeading(gs.GetLeading());
                                                    break;
                                                case GState.GStateAttribute.e_horizontal_scale:
                                                    ngs.SetHorizontalScale(gs.GetHorizontalScale());
                                                    break;
                                                case GState.GStateAttribute.e_word_spacing:
                                                    ngs.SetWordSpacing(gs.GetWordSpacing());
                                                    break;
                                                case GState.GStateAttribute.e_char_spacing:
                                                    ngs.SetCharSpacing(gs.GetCharSpacing());
                                                    break;
                                                case GState.GStateAttribute.e_dash_pattern:
                                                    ngs.SetDashPattern(gs.GetDashes(), gs.GetPhase());
                                                    break;
                                                case GState.GStateAttribute.e_miter_limit:
                                                    ngs.SetMiterLimit(gs.GetMiterLimit());
                                                    break;
                                                case GState.GStateAttribute.e_flatness:
                                                    ngs.SetFlatness(gs.GetFlatness());
                                                    break;
                                                case GState.GStateAttribute.e_line_join:
                                                    ngs.SetLineJoin(gs.GetLineJoin());
                                                    break;
                                                case GState.GStateAttribute.e_line_cap:
                                                    ngs.SetLineCap(gs.GetLineCap());
                                                    break;
                                                case GState.GStateAttribute.e_line_width:
                                                    ngs.SetLineWidth(gs.GetLineWidth());
                                                    break;
                                                case GState.GStateAttribute.e_fill_color:
                                                    ngs.SetFillColor(gs.GetFillColor());
                                                    break;
                                                case GState.GStateAttribute.e_fill_cs:
                                                    ngs.SetFillColorSpace(gs.GetFillColorSpace());
                                                    break;
                                                case GState.GStateAttribute.e_stroke_color:
                                                    ngs.SetStrokeColor(gs.GetStrokeColor());
                                                    break;
                                                case GState.GStateAttribute.e_stroke_cs:
                                                    ngs.SetStrokeColorSpace(gs.GetStrokeColorSpace());
                                                    break;
                                                case GState.GStateAttribute.e_rendering_intent:
                                                    ngs.SetRenderingIntent(gs.GetRenderingIntent());
                                                    break;
                                                case GState.GStateAttribute.e_transform:
                                                    ngs.SetTransform(gs.GetTransform());
                                                    break;
                                            }
                                        }

                                        ngs.SetFillColorSpace(ColorSpace.CreateDeviceRGB());

                                        if (gbox.IntersectRect(gbox, sel_rect))
                                        {
                                            ngs.SetFillColor(textColor);
                                        }
                                        else
                                        {
                                            ngs.SetFillColor(cpt);
                                        }

                                        writer.WriteElement(new_element);
                                        gs.SetFillColor(cpt);
                                    }
                                }
                            }
                            break;
                        }
					case Element.Type.e_form:
						{
							writer.WriteElement(element); // write Form XObject reference to current stream

                            Obj form_obj = element.GetXObject();
							if (!visited.Contains(form_obj.GetObjNum())) // if this XObject has not been processed
							{
								// recursively process the Form XObject
								visited.Add(form_obj.GetObjNum());
								ElementWriter new_writer = new ElementWriter();

								reader.FormBegin();
								new_writer.Begin(form_obj, true);
								ProcessElements(reader, new_writer, visited);
								new_writer.End();
								reader.End();
							}
							break;
						}
                    default:
                        writer.WriteElement(element);
                        break;
				}
			}
		}

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			PDFNet.Initialize();

			// Relative path to the folder containing test files.
			string input_path = "../../../../TestFiles/";
			string output_path = "../../../../TestFiles/Output/";
			string input_filename = "newsletter.pdf";
			string output_filename = "newsletter_edited.pdf";

			try
			{
				Console.WriteLine("Opening the input file...");
				using (PDFDoc doc = new PDFDoc(@"D:\work\PDFNet_alpha\CWrap\Samples\TestFiles\newsletter.pdf"))
				{
					doc.InitSecurityHandler();

					ElementWriter writer = new ElementWriter();
					ElementReader reader = new ElementReader();
					XSet visited = new XSet();

					PageIterator itr = doc.GetPageIterator();

					while (itr.HasNext())
					{
						Page page = itr.Current();
						visited.Add(page.GetSDFObj().GetObjNum());

						reader.Begin(page);
						writer.Begin(page, ElementWriter.WriteMode.e_replacement, false);
						ProcessElements(reader, writer, visited);
						writer.End();
						reader.End();

						itr.Next();
                        break;
					}

					doc.Save(@"D:\work\PDFNet_alpha\CWrap\Samples\TestFiles\newsletter_out_04.pdf", SDFDoc.SaveOptions.e_remove_unused);
					Console.WriteLine("Done. Result saved in {0}...", output_filename);
				}
			}
			catch (PDFNetException e)
			{
				Console.WriteLine(e.Message);
			}

		}
	}
}
